#############################################################################
# pyvox.py - Python extension module for Voxel Kit bindings
#
# Author:  Paul Hughett
#
# Copyright (c) 2001 Trustees of the University of Pennsylvania
# See the file COPYING for license terms.
#
#
# This file contains the Python-language part of the pyvox extension to
# Python, and imports the appropriate objects from the C language part.
#
# FIXME: Provide a way of getting the error object created by pyvoxC
#
# FIXME: Generalize here or in mmul to allow rank-1 vectors to be treated
# as row or column vectors as appropriate.
#
# FIXME: Check for proper shape and type of arguments.  Throw an exception
# on mismatch?
#
# FIXME: Remove hello and affine.hello, which were needed only to test
# the way to do upcalls?  Perhaps wait until there are some other good
# examples that also do something useful?
#
# $Id: pyvox.py,v 1.20 2002/02/05 15:49:26 hughett Exp $ Release $Name:  $
#############################################################################


import exceptions
import os
import re
import stat
import tempfile

from math import sin, cos
from exceptions import Exception
from types import IntType, FloatType, ListType, TupleType
import exim
import pyvoxC


#--------------------------------------------------------------------------
# Import some specific symbols from the C language part of the extension
#--------------------------------------------------------------------------

from pyvoxC import array, const, kernel, ramp
from pyvoxC import ArrayType, KernelType



#--------------------------------------------------------------------------
# Pass pyvox dictionary to pyvox C
#
# This dictionary is stored within PyvoxC and permits the C level of pyvox 
# to call functions and methods defined in this Python module.
#--------------------------------------------------------------------------

pyvoxC.set_pyvox_dict(globals())



#------------------------------------------------------------------------
# pyvox.Error - Exception object to report errors
#
# FIXME: Pyvox and PyvoxC should use the same exception class to report
# errors.
#------------------------------------------------------------------------

class Error(exceptions.Exception) : 
    def __init__(self, args=None) :
        self.args = args



#-------------------------------------------------------------------------
# hello - test function of upcalls
#
# This function is used just to test the correctness of method used to
# call a function on the python level (i.e. here) from a function on the
# C level (i.e. upcall_function in pyvox.c).
#-------------------------------------------------------------------------

def hello(name=None) :
    if name :
        print "Hello, %s!" % name
    else :
        print "Hello, world!"
    return 42            # Thanks for all the fish!



#-------------------------------------------------------------------------
# Conversions to special types of Pyvox arrays
#
# These functions are made available to the Pyvox user to convert generic
# array representations into the specific representations assumed by
# particular functions within Pyvox.  They are also used within Pyvox
# itself to convert generic representations to specific ones.  These
# functions may also be used to check the size of the converted data
# and will throw an exception if the size is wrong; if no expected size
# is specified, then the size is determined by the data provided.
#
# If the data is already in the correct format, it is NOT copied but is
# returned just as it is.  (Except for diag, which always creates a new
# array.)
#-------------------------------------------------------------------------


# Convert to rank-1 vector (n) of type double
def vector(data, n=None) :
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n] and \
               data.type == exim.double : 
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.double().reshape([n])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([n], exim.double, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([n], exim.double, data)
            elif isinstance(data, IntType) or isinstance(data, FloatType) :
                return array([n], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %d-element numeric vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 1 and \
               data.type == exim.double:
                return data
            elif isinstance(data, ArrayType) :
                return data.double().reshape([len(data)])
            elif isinstance(data, TupleType) :
                return array([len(data)], exim.double, data)
            elif isinstance(data, ListType) :
                return array([len(data)], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric vector" % repr(data)
            raise Error, msg


# Convert to row vector (1 * n) of type double
def row(data, n=None) :
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [1, n] \
               and data.type == exim.double :
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.double().reshape([1,n])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([1,n], exim.double, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([1,n], exim.double, data)
            elif isinstance(data, IntType) or isinstance(data, FloatType) :
                return array([1,n], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %d-element numeric column vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 2 \
               and data.size[0] == 1 and data.type == exim.double :
                return data
            elif isinstance(data, ArrayType) :
                return data.double().reshape([1, len(data)])
            elif isinstance(data, TupleType) :
                return array([1, len(data)], exim.double, data)
            elif isinstance(data, ListType) :
                return array([1, len(data)], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric column vector" % repr(data)
            raise Error, msg


# Convert to column vector (n * 1) of type double
def column(data, n=None) :
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n,1] \
               and data.type == exim.double :
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.double().reshape([n,1])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([n,1], exim.double, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([n,1], exim.double, data)
            elif isinstance(data, IntType) or isinstance(data, FloatType) :
                return array([n,1], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %d-element numeric column vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 2 \
               and data.size[1] == 1 and data.type == exim.double :
                return data
            elif isinstance(data, ArrayType) :
                return data.double().reshape([len(data), 1])
            elif isinstance(data, TupleType) :
                return array([len(data), 1], exim.double, data)
            elif isinstance(data, ListType) :
                return array([len(data), 1], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric column vector" % repr(data)
            raise Error, msg


# Convert to coordinate point
# Just an alias for column() but clearer to user in many contexts
point = column


# Convert to diagonal matrix (n by n) of type double
def diag(data, n=None) :
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n] and \
               data.type == exim.double : 
                vector = data
            elif isinstance(data, ArrayType) and len(data) == n :
                vector = data.double().reshape([n])
            elif isinstance(data, TupleType) and len(data) == n :
                vector = array([n], exim.double, data)
            elif isinstance(data, ListType) and len(data) == n :
                vector = array([n], exim.double, data)
            elif isinstance(data, IntType) or isinstance(data, FloatType) :
                vector = array([n], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %d-dimensional diagonal matrix" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 1 and \
               data.type == exim.double:
                vector = data
            elif isinstance(data, ArrayType) :
                vector = data.double().reshape([len(data)])
            elif isinstance(data, TupleType) :
                vector = array([len(data)], exim.double, data)
            elif isinstance(data, ListType) :
                vector = array([len(data)], exim.double, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a diagonal matrix" % repr(data)
            raise Error, msg
    n = len(vector)
    result = array([n,n], exim.double, 0)
    for i in range(n) :
        result[i,i] = vector[i]
    return result


# Convert to a rank-2 matrix; returns *diagonal* matrix for scalar data
# FIXME: Allow data to be a flat list if (nr, nc) specified?
# FIXME: Allow nr to be the list (nr, nc)?
def matrix(data, nr=None, nc=None) :
    if nr and not nc :
        nc = nr
    try :
        if isinstance(data, ArrayType) :
            if nr and data.size == [nr,nc] and data.type == exim.double :
                return data
            elif nr and len(data) == nr*nc :
                return data.double().reshape([nr,nc])
            elif not nr and data.rank == 2 :
                return data.double()
            else :
                raise Error
        elif isinstance(data, IntType) or isinstance(data, FloatType) :
            if not nr :
                raise Error
            else :
                result = array([nr, nc], exim.double, 0)
                if nr < nc : nn = nr
                else:        nn = nc
                for n in range(nn):
                    result[n,n] = data
                return result
        elif isinstance(data, ListType ) or isinstance(data, TupleType) :
            nr1 = len(data)
            nc1 = len(data[0])
            for r in range(nr1) :
                if len(data[r]) != nc1 :
                    raise Error
            if (nr and nr != nr1) or (nc and nc != nc1) :
                raise Error
            result = array([nr1, nc1], exim.double, 0)
            for r in range(nr1) :
                for c in range(nc1) :
                    result[r,c] = data[r][c]
            return result
        else :
            raise Error
    except :
        if nr :
            msg = "Cannot convert %s to a %d by %d numeric matrix" % \
                  (repr(data), nr, nc)
        else :
            msg = "Cannot convert %s to a numeric matrix" % \
                  repr(data)
        raise Error, msg


                        
#--------------------------------------------------------------------------
# Type checking and conversion
#
# If x can be converted to type foo, do so and return it; else raise an
# exception.  Unadvertized and intended for internal use.
#
# FIXME: These may eventually get replaced by the conversion functions
# given just above.
#--------------------------------------------------------------------------


# Check if x can reasonably be interpreted as an integer
def check_int(x):
    if isinstance(x, IntType):
        return x
    elif isinstance(x, FloatType):
        return int(x)
    else:
        raise Error, 'Cannot interpret arg as an integer'


# Check if x can reasonably be interpreted as a float value
def check_float(x):
    if isinstance(x, FloatType):
        return x
    elif isinstance(x, IntType):
        return float(x)
    else:
        raise Error, 'Cannot interpret arg as a float value'


# Check if x can reasonably be interpreted as a rank-1 vector of n elements
# (A scalar will be interpreted as a constant vector.)
def check_vector(x, n) :
    if isinstance(x, ArrayType) and x.size == [n] :
        return x
    if isinstance(x, ArrayType) and x.size == [1, n] :
        return x.double().reshape([n])
    try :
        return array([n], exim.double, x)
    except :
        raise Error, "Cannot interpret as a numeric vector"


# Check if x can reasonably be interpreted as an n x 1 column vector
# (A scalar will be interpreted as a constant vector.)
def check_column(x, n) :
    if isinstance(x, ArrayType) and x.size == [n,1] :
        return x
    if isinstance(x, ArrayType) and x.size == [n] :
        return x.double().reshape([n,1])
    try :
        return array([n,1], exim.double, x)
    except :
        raise Error, 'Cannot interpret as an n x 1 column vector'


# Check if x can reasonably be interpreted as an nr x nc matrix
def check_matrix(x, nr, nc) :
    if isinstance(x, ArrayType) and x.size == [nr, nc] :
        return x
    try :
        return array([nr,nc], exim.double, x)
    except :
        raise Error, 'Cannot interpret as a matrix of right size'



#--------------------------------------------------------------------------
# pyvox.monomials - Compute monomials of degree <= n in k literals
#
# This function lists the monomials of total degree less than or equal to
# n in k literals, as an array of k columns by the number of distinct
# monomials; each row of the table gives the power to be applied to the
# corresponding literal.  For example, evaluating monomials(2,2) might
# give the following array and corresponding monomials (with literals y,x):
#
#       0   0    =>  1
#       0   1    =>  x
#       0   2    =>  x^2
#       1   0    =>  y
#       1   1    =>  xy
#       2   0    =>  y^2
#
# The order of the rows in the table is not guaranteed to be consistent
# from one version of pyvox to another.
#
# One application for this function is to list the moments of order n or
# less in k dimensions, and to provide a defined order for listing the
# moments in a generic moment calculating function.
#--------------------------------------------------------------------------

def monorec(n, k, prefix) :

    if k == 0 :
        return prefix
    elif n == 0 :
        return prefix + k*[0,]
    else :
        ans = []
        for p in range(n+1) :
            ans = ans + monorec(n-p, k-1, prefix+[p,])
        return ans


def monomials(n, k) :

    # Return None if n or k is too small
    if n < 0 or k < 0:
        return None

    # Call recursive function to compute the monomials
    ans = monorec(n, k, [])

    # Reformat as a pyvox array
    nrows = len(ans) / k
    ans = array([nrows, k], exim.int, ans)
    return ans



#-------------------------------------------------------------------------
# rawread - Read a possibly compressed raw image file
#
# This function reads an image file of some given name, type, and size 
# into a Pyvox array, and returns the array.  If the given filename
# ends in ".gz" or ".Z", it is assumed to be a compressed file and is
# uncompressed into a temporary file and then read. If there is no image
# file with the given name, but there is a file with either ".gz" or ".Z"
# appended, then that file is assumed to be the image file compressed,
# which is uncompressed on the fly and read instead.  The first element
# of the dimen array may be zero, in which case the number of slices
# is determined by the size of the (uncompressed) image file.
#
# FIXME: Check during ./configure whether or not the current platform
# support gunzip or uncompress.
#
# FIXME: Add support for bzip2?
#--------------------------------------------------------------------------

def rawread(filename, dimen, extype=exim.uint1, bigend=1) :

    # Check first for an uncompressed file; then for a compressed file
    if os.access(filename, os.R_OK) :
        if re.search(r'\.gz$', filename) or re.search(r'\.Z$', filename) :
            gzfile = filename
        else :
            gzfile = None
    elif os.access(filename+".gz", os.R_OK) :
        gzfile = filename + ".gz"
    elif os.access(filename+".Z",  os.R_OK) :
        gzfile = filename + ".Z"
    else :
        raise Exception, "Unable to read image file %s" % (filename,)

    # Uncompress the file into a temporary, if necessary
    if gzfile :
        tmpfile = tempfile.mktemp('.img')
        imgfile = tmpfile
        cmd = "gunzip -c %s > %s" % (gzfile, tmpfile)
        os.system(cmd)
    else :
        tmpfile = None
        imgfile = filename

    # Determine number of bytes per pixel
    # FIXME: Define a function in exim to do this
    if extype == exim.uint1 or extype == exim.int1 :
        nbytes = 1
    elif extype == exim.uint2 or extype == exim.int2 :
        nbytes = 2
    else :
        raise Exception, "Unable to interpret extype %i" % (extype,)

    # If necessary, determine number of slices from size of the file
    if dimen[0] == 0 :
        npix = 1          # Pixels per slice
        for i in range(len(dimen)-1) :
            npix = npix * dimen[i+1]
        imgstat = os.stat(imgfile)
        nslc = imgstat[stat.ST_SIZE] / (nbytes * npix)
        dim2 = [nslc,] + list(dimen[1:])
    else :
        dim2 = list(dimen)

    # Actually read the file 
    image = pyvoxC.read(imgfile, dim2, extype, bigend)

    # Remove the temporary file, if used
    if tmpfile :
        os.remove(tmpfile)

    # Return the array object
    return image



#---------------------------------------------------------------------------
# pyvox.read - Deprecated for pyvox.rawread
#---------------------------------------------------------------------------

def read(filename, dimen, fmt=None, bigend=None) :
    raise Exception, "pyvox.read is deprecated; use pyvox.rawread instead"



#----------------------------------------------------------------------------
# pyvox.bblanz - Contents of a BBL analyze header
#
# An instance of this class contains the contents of a bblanz header read
# when the instance is initialized.  Attributes of the instance contain the
# values of selected fields; the selection is more or less defined by
# which fields actually contain useful values at BBL.  The attributes
# currently defined include:
#
#    hdrfile    Path and name of the header file
#    imgfile    Path and name of the image file
#    rank       Number of dimensions in image
#    size       Number of pixels along each dimension, in pyvox order
#    spacing    Physical spacing along each dimension, in pyvox order
#    extype     External type of image data
#    bigend     Bigendian?
#
# FIXME: This is sufficiently BBL-centric that it ought to be moved into
# the nria or bwarp modules, along with the CLI programs that access this
# header format.
#---------------------------------------------------------------------------

class bblanz :

    # Initialize an instance
    # All the real work in done at the C level.
    def __init__(self, file) :
        pyvoxC.init_bblanz(self, file)



#---------------------------------------------------------------------------
# pyvox.affine - Affine transform
#
# This class allows the representation and manipulation of affine
# transforms in any number of dimensions.  With a few exceptions,
# operations such as translate, scale, and rotate operate in place
# rather than creating a new transform.
#
# In compositions, the order of composition can be selected with the
# optional pre argument, which takes the following possible values:
#
#   1 => precompose  (first position)
#   2 => postcompose (second position)
#   0 => postcompose (for compatibility only; pre used to be a boolean
#        meaning precompose)
#  -1 => precompose with inverse
#  -2 => postcompose with inverse
#
# NOTE: To avoid a conflict with the matrix() function, we use matrox
# as the name of the local variable containing the matrix part of the
# affine transform.
#
# FIXME: Add repr() and str() methods.  Add a prettyprint method.
# FIXME: Add method to return index to params.  Arrange for this class and
# afflst to use the same parameter order.
#---------------------------------------------------------------------------

class affine:
    "Affine transform"

    # Initialize an affine transform to given values: 
    # A single scalar argument yields the identity transform for
    # that number of dimensions.
    def __init__(self, ndim, matrox=None, offset=None):
        self.ndim = ndim
        if matrox :
            self.matrix = matrix(matrox, ndim, ndim)
        else :
            self.matrix = array([ndim,ndim], exim.double, 0)
            for i in range(ndim):
                self.matrix[i,i] = 1.0
        if offset :
            self.offset = point(offset, ndim)
        else :
            self.offset = array([ndim,1], exim.double, 0)

    # Print self in readable format
    # FIXME: Add a stream to print to
    def write(self):
        for i in range(self.ndim):
            for j in range(self.ndim):
                print self.matrix[i,j], 
            print self.offset[i,0]
        print

    # Deep copy
    def copy(self) :
        return affine(self.ndim, self.matrix.copy(), self.offset.copy())

    # Return a vector of parameters defining the transform
    # For compatibility with optimizers of various flavors
    def param(self) :
        ndim   = self.ndim
        nparm  = ndim*(ndim+1)
        result = array(nparm, exim.double, 0)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                result[iparm] = self.matrix[irow, icol]
            result[irow*(ndim+1)+ndim] = self.offset[irow, 0]
        return result

    # Initialize from a vector of parmeters defining the transform
    # For compatibility with optimizers of various flavors
    def setparam(self, parm) :
        ndim  = self.ndim
        nparm = ndim*(ndim+1)
        parm  = vector(parm, nparm)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                self.matrix[irow, icol] = parm[iparm]
            self.offset[irow, 0] = parm[irow*(ndim+1)+ndim]
        return self

    # Add an increment to the parameters defining the transform
    # For compatibility with optimizers of various flavors
    def addparam(self, parm) :
        ndim  = self.ndim
        nparm = ndim*(ndim+1)
        parm  = vector(parm, nparm)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                self.matrix[irow, icol] = self.matrix[irow, icol] + parm[iparm]
            ioff = irow*(ndim+1)+ndim
            self.offset[irow, 0] = self.offset[irow, 0] + parm[ioff]
        return self

    # Compose with another affine transform
    def compose(self, other, pre=0):
        if not isinstance(other, affine) :
            raise Error, 'Other is not an affine transform'
        if self.ndim != other.ndim :
            raise Error, 'Self and other have different number of dimensions'
        if pre == 0 or pre == 2 :
            self.offset = other.matrix.mmul(self.offset) + other.offset
            self.matrix = other.matrix.mmul(self.matrix)
        elif pre == 1 :
            self.offset = self.matrix.mmul(other.offset) + self.offset
            self.matrix = self.matrix.mmul(other.matrix)
        elif pre == -1 :
            matrox = other.matrix.inverse()
            offset = matrox.mmul(-other.offset)
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -2 :
            matrox = other.matrix.inverse()
            self.offset = matrox.mmul(self.offset - other.offset)
            self.matrix = matrox.mmul(self.matrix)
        else :
            raise Exception, "Invalid value of pre"
        return self

    # Compose with given matrix and offset of an affine transform
    def compose2(self, matrox, offset, pre=0):
        matrox = matrix(matrox, self.ndim, self.ndim)
        offset = point(offset, self.ndim)
        if pre == 0 or pre == 2 :
            self.offset = matrox.mmul(self.offset) + offset
            self.matrix = matrox.mmul(self.matrix)
        elif pre == 1 :
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -1 :
            matrox = matrox.inverse()
            offset = matrox.mmul(-offset)
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -2 :
            matrox = matrox.inverse()
            self.offset = matrox.mmul(self.offset - offset)
            self.matrix = matrox.mmul(self.matrix)
        else :
            raise Exception, "Invalid value of pre"            
        return self

    # Compose with matrix of an affine transform
    def composematrix(self, matrox, pre=0):
        matrox = matrix(matrox, self.ndim, self.ndim)
        if pre == 0 or pre == 2 :
            self.offset = matrox.mmul(self.offset)
            self.matrix = matrox.mmul(self.matrix)
        elif pre == 1:
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -1 :
            self.matrix = self.matrix.mmul(matrox.inverse())
        elif pre == -2 :
            matrox = matrox.inverse()
            self.offset = matrox.mmul(self.offset)
            self.matrix = matrox.mmul(self.matrix)
        else:
            raise Exception, "Invalid value of pre"  
        return self

    # Compose with index-to-physical transform given by an origin and scale
    def i2p(self, origin, scale, pre=0) :
        matrox = self.scalematrix(scale)
        offset = point(origin, self.ndim)
        return self.compose2(matrox, offset, pre)

    # Compose with physical-to-index transform given by an origin and scale
    def p2i(self, origin, scale, pre=0) :
        matrox = self.scalematrix(scale)
        offset = point(origin, self.ndim)
        matrox = matrox.inverse()
        offset = matrox.mmul(-offset)
        return self.compose2(matrox, offset, pre)

    # Invert the transformation
    def invert(self):
        self.matrix = self.matrix.inverse()
        self.offset = self.matrix.mmul(-self.offset)
        return self

    # Compute new transform which is the inverse of the old
    def inverse(self):
        matrox = self.matrix.inverse()
        offset = matrox.mmul(-self.offset)
        return affine(self.ndim, matrox, offset)

    # Compose with a translation
    def translate(self, delta, pre=0):
        delta = point(delta, self.ndim)
        if pre == 0 or pre == 2:
            self.offset = self.offset + delta
        elif pre == 1 :
            self.offset = self.offset + self.matrix.mmul(delta)
        elif pre == -1 :
            self.offset = self.offset - self.matrix.mmul(delta)
        elif pre == -2 :
            self.offset = self.offset - delta
        else :
            raise Exception, "Invalid value of pre"
        return self

    # Generate scaling matrix from scalar or vector
    def scalematrix(self, scale):
        scale = vector(scale, self.ndim)
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
                matrox[i,i] = scale[i]
        return matrox

    # Compose with a scaling, isotropic or not
    def scale(self, scale, pre=0):
        matrox = self.scalematrix(scale)
        return self.composematrix(matrox, pre)

    # Generate an elementary rotation matrix
    def rotationmatrix(self, axis1, axis2, angle):
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
            matrox[i,i] = 1.0
        matrox[axis1, axis1] =  cos(angle)
        matrox[axis1, axis2] =  sin(angle)
        matrox[axis2, axis1] = -sin(angle)
        matrox[axis2, axis2] =  cos(angle)
        return matrox

    # Compose with an elementary rotation matrix
    def rotate(self, axis1, axis2, angle, pre=0):
        d = self.rotationmatrix(axis1, axis2, angle)
        return self.composematrix(d, pre)

    # Compose with a rotation around an axis in 3D
    def rotate3d(self, axis, angle, pre=0) :
        if self.ndim != 3 :
            raise Error, 'rotate3d only works in 3 dimensions'
        axis = point(axis, 3)
        q0 = cos(angle/2.0)
        q1 = axis[0,0] * sin(angle/2.0)
        q2 = axis[1,0] * sin(angle/2.0)
        q3 = axis[2,0] * sin(angle/2.0)
        matrox = pyvox.array([3,3], exim.double, 0)
        matrox[0,0] = q0**2 + q1**2 - q2**2 - q3**2
        matrox[0,1] = 2.0*(q1*q2 - q0*q3)
        matrox[0,2] = 2.0*(q1*q3 + q0*q2)
        matrox[1,0] = 2.0*(q1*q2 + q0*q3)
        matrox[1,1] = q0**2 + q2**2 - q1**2 - q3**2
        matrox[1,2] = 2.0*(q2*q3 - q0*q1)
        matrox[2,0] = 2.0*(q1*q3 - q0*q2)
        matrox[2,1] = 2.0*(q2*q3 + q0*q1)
        matrox[2,2] = q0**2 + q3**2 - q1**2 - q2**2
        return self.composematrix(matrox, pre)

    # Generate an elementary shear matrix
    def shearmatrix(self, axis1, axis2, coef):
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
            matrox[i,i] = 1.0
        matrox[axis1, axis2] =  coef
        return matrox

    # Compose with an elementary shear matrix
    def shear(self, axis1, axis2, coef, pre=0):
        d = self.shearmatrix(axis1, axis2, coef)
        return self.composematrix(d, pre)

    # Transform and resample with linear interpolation
    # (Fast implementation)
    def linear(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        return pyvoxC.affine_linear(matrox, offset, source, dimen)
            
    # Transform and resample with linear interpolation
    # (Reference implementation)
    def linear0(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        return pyvoxC.affine_linear0(matrox, offset, source, dimen)

    # Transform and resample with nearest neighbor interpolation
    # (Fast implementation)
    def nearest(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        return pyvoxC.affine_nearest(matrox, offset, source, dimen)
    
    # Transform and resample with nearest neighbor interpolation
    # (Reference implementation)
    def nearest0(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        return pyvoxC.affine_nearest0(matrox, offset, source, dimen)
    
    # Transform a point
    # FIXME: Use plain call to transform a point?
    def point(self, x):
        x = point(x, self.ndim)
        return self.matrix.mmul(x) + self.offset

    # Transform a vector
    def vector(self, x):
        x = point(x, self.ndim)
        return self.matrix.mmul(x)

    # Test how to call a Python function from the C level
    def hello(self, name=None) :
        if name :
            print "Hello, %s!" % name
        else :
            print "Hello, world!"
        return 42            # Thanks for all the fish!

    # Compute a norm on affine transforms
    def norm(self, other=None, length=100.0) :
        ndim = self.ndim
        if other == None or other == 0 :
            matrox = self.matrix
            offset = self.offset
        elif other == 1 :
            matrox = self.matrix
            offset = self.offset
            for n in range(ndim) :
                matrox[n,n] = matrox[n,n] - 1.0
        else :
            matrox = self.matrix - other.matrix
            offset = self.offset - other.offset
        result = 0.0
        for r in range(ndim) :
            res = 0.0
            for c in range(ndim) :
                res = res + abs(matrox[r,c])
            res = length*res + abs(offset[r,0])
            if res > result :
                result = res
        return result
    
# Provide a class object similar to the other Pyvox type objects
AffineType = affine



#---------------------------------------------------------------------------
# spline - Cubic spline transform
#
# This class allows the definition of coordinate transforms (or other
# functions) defined by cubic splines.
#
# Instance variables are: nin = number of dimensions for input; nout =
# number of dimensions for output; ngrid = number of nodes in grid;
# intrans = affine transform for input; outtrans = affine transform for
# output; offset = array of affine offsets; matrix = array of affine matrices.
#
# FIXME: Incomplete; this is just a tentative outline of the interface.
#---------------------------------------------------------------------------

class spline :
    "Cubic spline transform"

    # Initialize a cubic spline transform
    def __init__(self, nin, nout, ngrid) :
        self.nin   = nin
        self.nout  = nout
        self.ngrid = ngrid
        self.intrans  = affine(nin)
        self.outtrans = affine(nout)
        # FIXME: Decide format and initial values for offset and matrix

    # Evaluate at a point
    def point(self, x) :
        pass     # FIXME: Not yet implemented

    # Return affine approximation at a point
    def local(self, x) :
        pass     # FIXME: Not yet implemented

    # Set the local affine approximation at at node
    def setlocal(self, node, affine) :
        pass     # FIXME: Not yet implemented

    # Compose with an affine transform
    def compose(self, affine, pre=1) :
        pass     # FIXME: Not yet implemented

    # Resample an image using nearest neighbor interpolation
    def nearest(self, source, dimen) :
        pass     # FIXME: Not yet implemented

    # Resample an image using linear interpolation
    def linear(self, source, dimen) :
        pass     # FIXME: Not yet implemented
